#include <utility>
#include <cstdio>
#include <vector>
#include <queue>
#include <map>
#include <fcntl.h>
#include <sys/poll.h>
#include <csignal>

#include "buffer.hh"

using std::make_pair;

static unsigned total_bytes = 0;

class buffer
{
    std::queue<BufferType> buffers;
    unsigned frontptr;
public:
    buffer() : frontptr(0)
    {
    }
    buffer(const BufferType& data) : frontptr(0)
    {
        PushBack(data);
    }
    
    void PushBack(const BufferType& data)
    {
     /* Don't use string append - it's too slow and we don't need it. */
        buffers.push(data);
    }
    void PopFront(unsigned n)
    {
        while(n > 0 && !buffers.empty())
        {
            unsigned c = GetFrontSize();
            if(n < c)
            {
                frontptr += n;
                c = n; // ate "n" bytes
            }
            else
            {
                buffers.pop();
                frontptr = 0;
                // ate everything that was left in the front buf
            }
            n -= c;
        }
    }
    unsigned GetFrontSize() const
    {
        if(buffers.empty()) return 0;
        return buffers.front().size() - frontptr;
    }
    const char *GetFrontPtr() const
    {
        return &buffers.front()[0] + frontptr;
    }
    
    bool Empty() const
    {
        return GetFrontSize() == 0;
    }
    
    unsigned GetBufSize() const
    {
        return Empty()
             ? 0
             : buffers.front().size() - frontptr
             + buffers.back().size() * (buffers.size() - 1);
    }
};

static void BlockSignals()
{
    sigset_t set;
    sigemptyset(&set);
    sigaddset(&set, SIGALRM);
    sigprocmask(SIG_BLOCK, &set, NULL);
}

static void UnblockSignals()
{
    sigset_t set;
    sigemptyset(&set);
    sigaddset(&set, SIGALRM);
    sigprocmask(SIG_UNBLOCK, &set, NULL);
}

static std::map<int, buffer> buffers;
static std::map<int, FILE*> fileps;

static void UpdateBuffer(int fd, buffer& buf)
{
    unsigned count = buf.GetFrontSize();
    if(!count) return;

    const char *p = buf.GetFrontPtr();
    
    int n = write(fd, p, count);
    if(n <= 0) return;
    
    buf.PopFront(n);
    // retry!
    UpdateBuffer(fd, buf);
}

void AddWriteTask(FILE *fp, const BufferType& data)
{
    BlockSignals();
    
    int fd = fileno(fp);
    
    total_bytes += data.size();
    
    std::map<int, buffer>::iterator i;
    
    i = buffers.lower_bound(fd);
    if(i == buffers.end() || i->first != fd)
    {
        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
        
        buffer tmp = data;
        UpdateBuffer(fd, tmp);
        
        if(tmp.Empty())
        {
            BufferEmptied(fp);
            
            UnblockSignals();
            return;
        }
        
        i = buffers.insert(i, make_pair(fd, tmp));
        fileps.insert(make_pair(fd, fp));
    }
    else
    {
        i->second.PushBack(data);
    
        UpdateBuffer(fd, i->second);
    }
    
    UnblockSignals();
}

static void DeleteBuffer(std::map<int, buffer>::iterator i)
{
    std::map<int, FILE*>::iterator fpi = fileps.find(i->first);

    BufferEmptied(fpi->second);
    
    buffers.erase(i);
    fileps.erase(fpi);
}

void UpdateBuffers()
{
    std::map<int, buffer>::iterator i, j;
    for(i = buffers.begin(); i != buffers.end(); i=j)
    {
        j=i; ++j;
        
        UpdateBuffer(i->first, i->second);
        
        if(i->second.Empty())
        {
            DeleteBuffer(i);
        }
    }
}

void FlushBuffers()
{
    std::map<int, buffer>::iterator i,j;
    for(i = buffers.begin(); i != buffers.end(); i=j)
    {
        j=i; ++j;
        
        int fd = i->first;

        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) &~ O_NONBLOCK);

        buffer& buf = i->second;
        
        for(;;)
        {
            unsigned count = buf.GetFrontSize();
            if(!count) break;
            
            const char *p = buf.GetFrontPtr();
            
            int n = write(fd, p, count);
            if(n > 0)
            {
                buf.PopFront(n);
            }
        }
        
        DeleteBuffer(i);
    }
}

const string GetBufferStatus()
{
    unsigned num_bytes = 0;
    
    std::map<int, buffer>::const_iterator i;
    for(i = buffers.begin(); i != buffers.end(); ++i)
    {
        num_bytes += i->second.GetBufSize();
    }
    
    char Buf[64];
    sprintf(Buf, "%uk/%uM", num_bytes / 1024, total_bytes/1048576);
    return Buf;
}

bool BufferElapse(double nseconds)
{
    std::vector<pollfd> polls;

    std::map<int, buffer>::iterator i;
    for(i = buffers.begin(); i != buffers.end(); ++i)
    {
        if(i->second.Empty()) continue;
        
        pollfd tmp = { i->first, POLLOUT, 0 };
        polls.push_back(tmp);
    }
    
    if(polls.empty())
    {
        usleep((unsigned)(nseconds * 1e6));
        return false;
    }
    
    int timeout = (int)(nseconds * 1000);
    if(!timeout) timeout=1;
    
    int res = poll(&polls[0], polls.size(), timeout);
    if(!res) return true;
    
    /* Block the timer signal while writing so that integrity won't fail */
    BlockSignals();
    
    for(unsigned a=0; a<polls.size(); ++a)
        if(polls[a].revents & POLLOUT)
        {
            int fd = polls[a].fd;

            i = buffers.find(fd);
            UpdateBuffer(fd, i->second);
            
            if(i->second.Empty())
            {
                DeleteBuffer(i);
            }
        }
    
    UnblockSignals();

    return false;
}

bool BufferWaitInput()
{
    for(;;)
    {
        std::vector<pollfd> polls;

        std::map<int, buffer>::iterator i;
        for(i = buffers.begin(); i != buffers.end(); ++i)
        {
            if(i->second.Empty()) continue;
            
            pollfd tmp = { i->first, POLLOUT, 0 };
            polls.push_back(tmp);
        }
        
        pollfd tmp = { STDIN_FILENO, POLLIN, 0 };
        polls.push_back(tmp);
        
        int status = poll(&polls[0], polls.size(), -1);
        if(status < 0) break; // error or interruption

        /* Block the timer signal while writing so that integrity won't fail */
        BlockSignals();
        
        for(unsigned a=0; a<polls.size(); ++a)
        {
            if(polls[a].revents & POLLOUT)
            {
                int fd = polls[a].fd;
                
                i = buffers.find(fd);
                UpdateBuffer(fd, i->second);
                
                if(i->second.Empty())
                {
                    DeleteBuffer(i);
                }
            }
            if(polls[a].revents & POLLIN)
            {
                return true;
            }
        }
        
        UnblockSignals();
    }
    return false;
}

BufferType::BufferType(unsigned size, char c)
{
    data = new BufferData;
    data->data = new char[size];
    data->size = size;
    std::fill_n(data->data, size, c);
}
BufferType::BufferType(unsigned size)
{
    data = new BufferData;
    data->data = new char[size];
    data->size = size;
}
BufferData::~BufferData()
{
    delete[] data;
}
BufferType::BufferType(const char*b, const char*e)
{
    unsigned size = e-b;
    data = new BufferData;
    data->data = new char[size];
    data->size = size;
    std::copy(b,e, data->data);
}
BufferType::BufferType(const unsigned char*b, const unsigned char*e)
{
    unsigned size = e-b;
    data = new BufferData;
    data->data = new char[size];
    data->size = size;
    std::copy(b,e, data->data);
}
